private import std.stream;
private import std.string;
private import std.conv;
private import std.ctype;
private import std.regexp;
private import std.c.stdio;
private import std.c.stdlib;

extern (C) {
    typedef void (*regsym_cb) (char*, void*);
    int reflection_init(regsym_cb fp, char* exe);
    int reflection_dlopen(regsym_cb fp, char* file, char* func);
}

version (Windows) {
    version = classDoubleUnderbar;
}
version (darwin) {
    version = classDoubleUnderbar;
    version = methodDoubleUnderbar;
}

private {
    version (classDoubleUnderbar) {
        char[] classStr = `__Class_`;
        char[] classDStr = `__Class__D`;
    }
    else {
        char[] classStr = `_Class_`;
        char[] classDStr = `_Class__D`;
    }
    version (methodDoubleUnderbar) {
        char[] methodStr = `__D`;
    }
    else {
        char[] methodStr = `_D`;
    }
}

/// utility function
private bool startWith(char[] s1, char[] s2) {
    return s1.length >= s2.length && s1[0..s2.length] == s2;
}

private class Demangle {
    this() {
    }

    char[][] demangleName(char[] sym) {
        int i = 0;
        return demangleName(sym, i);
    }

    char[][] demangleName(char[] sym, inout int i) {
        if (sym[i..sym.length].startWith(methodStr)) i += methodStr.length;

        char[][] names;
        while (i < sym.length && std.ctype.isdigit(sym[i])) {
            int num = getNum(sym, i);

            char[] now = sym[i..i+num];

            if (now == "__anonymous") {
            }
            else {
                // we cannot demangle template correctly!
/*
                int sep = rfind(now, "_");
                if (sep > 0) {
                    try {
                        int dummy = sep+1;
                        int prev = dummy;
                        char[][] types;
                        while (dummy != now.length) {
                            Type t = demangleType(now, dummy);
                            if (prev == dummy || t === null) goto TEMPLATE_END;
                            prev = dummy;
                            types ~= t.toString;
                        }
                        now = now[0..sep] ~ "!(" ~ join(types,`,`) ~ ")";
                    }
                    catch (Object o) {}
                }
            TEMPLATE_END:
*/
                names ~= now;
            }

            i += num;
        }

        return names;
    }

    Type demangleType(char[] sym, inout int i) {
        Type type = new Type();

        switch (sym[i]) {
        case 'A': { // array
            i++;
            type = demangleType(sym, i);
            type.qual |= Type.Qual.ARRAY;
            return type;
        }
        case 'G': { // static array
            i++;
            char[] num = getNumStr(sym, i);
            type = demangleType(sym, i);
            type.qual |= Type.Qual.SARRAY;
            return type;
        }
        case 'H': { // assoc array
            i++;
            type.assoc = demangleType(sym, i);
            type = demangleType(sym, i);
            type.qual |= Type.Qual.ASSOC;
            return type;
        }
        case 'P': { // pointer
            i++;
            type = demangleType(sym, i);
            type.qual |= Type.Qual.POINTER;
            return type;
        }
        case 'K': { // inout
            i++;
            type = demangleType(sym, i);
            type.qual |= Type.Qual.INOUT;
            return type;
        }
        case 'J': { // out
            i++;
            type = demangleType(sym, i);
            type.qual |= Type.Qual.OUT;
            return type;
        } 
        case 'R': { // reference
            assert(false);
            i++;
            type = demangleType(sym, i);
            type.qual = Type.Qual.REFERENCE;
            return type;
        }
        case 'I': { // @@@ident?
            assert(false);
        }
        case 'C': { // class
            type.kind = Type.Kind.KLASS;
            i++;
            char[][] t = demangleName(sym, i);
            // @@@
            type.klass = Class.forName(join(t, `.`));
            return type;
        }
        case 'S': { // struct
            type.kind = Type.Kind.UNKNOWN;
            i++;
            char[][] t = demangleName(sym, i);
            // @@@
            type.unknown = join(t, `.`);
            return type;
        }
        case 'E': { // enum
            type.kind = Type.Kind.UNKNOWN;
            i++;
            char[][] t = demangleName(sym, i);
            // @@@
            type.unknown = join(t, `.`);
            return type;
        }
        case 'T': { // typedef
            type.kind = Type.Kind.UNKNOWN;
            i++;
            char[][] t = demangleName(sym, i);
            // @@@
            type.unknown = join(t, `.`);
            return type;
        }
        case 'Y':
        case 'Z': { // separater
            return null;
        }
        case 'D': { // delegate
            i++;
            type = demangleType(sym, i);
            assert(type.kind == Type.Kind.FUNC);
            type.func.qual = Function.FQual.DELEGATE;
            return type;
        }
        case 'F':
        case 'U':
        case 'W':
        case 'V':
        /*case 'T':*/ { // (static) function
            type.kind = Type.Kind.FUNC;
            type.func = new Function();
            if (sym[i] == 'U') type.func.qual |= Function.FQual.STATIC;

            i++;
            Type[] args;
            bool nowRet = false;
            while (1) {
                Type t = demangleType(sym, i);

                if (nowRet) {
                    type.func.ret = t;
                    type.func.args = args;
                    return type;
                }
                if (i == sym.length) {
                    type.func.ret = t;
                    Type n = new Type;
                    n.kind = Type.Kind.BUILTIN;
                    n.builtin = new Builtin(Builtin.Type.ARGS);
                    args ~= n;
                    type.func.args = args;
                    return type;
                }

                if (t) args ~= t;

                if (sym[i] == 'Z') {
                    nowRet = true;
                    i++;
                }
                if (sym[i] == 'Y') {
                    i++;
                }
            }
            assert(false);
        }
        default:
            type.kind = Type.Kind.BUILTIN;
            try {
                if (std.ctype.isdigit(sym[i])) {
                    /* it is inside function, skip */
                    i++;
                    return null;
                }
                type.builtin = new Builtin(sym[i]);
            }
            catch (Error e) {
                throw new Error(sym ~ ": cannot demangle");
            }
            i++;
            return type;
        }
        return null;
    }

    Method demangleMethod(char[] sym) {
        int i = 0;
        return demangleMethod(sym, i);
    }

    Method demangleMethod(char[] sym, inout int i) {
        if (sym[i..sym.length].startWith(methodStr)) i += methodStr.length;

        char[][] names = demangleName(sym, i);

        char[] name = join(names, `.`);

        if (name == "") return null;

        assert(sym.length > i);

        Type type = demangleType(sym, i);
        if (!type) return null;

        Method ret = new Method();

        ret.fullname = name;
        ret.name = name;
        ret.type = type;

        int seps = names.length - 1;
        Class c = Class.forName(join(names[0..seps], `.`));
        if (c) {
            ret.klass = c;
            ret.name = names[seps];
            c.methods ~= ret;
        }

        return ret;
    }

private:
    bool issym(char c) {
        return isalnum(c) || c == '_';
    }

    int getNum(char[] str, inout int i) {
        return toInt(getNumStr(str, i));
    }
    char[] getNumStr(char[] str, inout int i) {
        char[] numStr = "";
        while (std.ctype.isdigit(str[i])) {
            numStr ~= str[i];
            i++;
        }
        return numStr;
    }

private:

}

extern (C) {
    typedef int (*CompareFunction)(void*,void*);

    void *bsearch(void* key, void* base, uint nmemb, uint size,
                  int (*compar)(char[], char[]));

    int dstrcmp_(char[]* sp1, char[]* sp2) {
        char[] s1 = *sp1;
        char[] s2 = *sp2;
        if (s1.length < s2.length) return -1;
        else if (s1.length > s2.length) return 1;
        for (int i = 0; i < s1.length; i++) {
            if (s1[i] < s2[i]) return -1;
            else if (s1[i] > s2[i]) return 1;
        }
        return 0;
    }
}

// all members are static
class Reflection {
public:
    static void init(char[] nmfile) {
        auto Demangle demangler = new Demangle();

        if (nmfile) {
            version (Windows) {
                loadMapFile(nmfile);
            }
            else {
                loadNmFile(nmfile);
            }
        }
        else {
            loadBFD();
        }
        syms = sym2addr.keys;
        qsort(&syms[0], syms.length, syms[0].sizeof,
              cast(CompareFunction)&dstrcmp_);

        initClass(demangler);
        initMethod(demangler);
    }

public:
    static void* getAddr(char[] sym) { return sym2addr[sym]; }

private:
    static void initClass(Demangle d) {
        foreach (char[] sym; syms) {
            if (sym.startWith(classStr)) {
                if (sym.startWith(classDStr)) {
                    // class in function.
                    continue;
                }
                char[][] seps =
                    d.demangleName(sym[classStr.length..sym.length]);
                Class c = new Class();
                c.name = join(seps, `.`);
                c.sym = sym;
                c.address = getAddr(sym);
                Class.classes[c.name] = c;
            }
        }
    }

    static void initMethod(Demangle d) {
        foreach (char[] sym; syms) {
            if (sym.startWith(methodStr)) {
                Method m = d.demangleMethod(sym);
                if (m is null) continue;
                m.sym = sym;
                m.address = getAddr(sym);
            }
        }
    }

    extern (C) static void cb(char* name, void* value) {
//        printf("[D] %x %s\n", value, name);
        sym2addr[std.string.toString(name).dup] = value;
    }

    static void loadBFD() {
        version (Windows) {
            classStr = `_Class_`;
            classDStr = `_Class__D`;
            methodStr = `D`;
        }
        reflection_init(&cb, null);
    }

    static void loadNmFile(char[] nmfile) {
        File f = new File(nmfile);
        char[] line;
        while ((line = f.readLine()) != null) {
            if (line[0] == ' ') continue;

            void* addr;
            sscanf(toStringz(line[0..8]), "%x", &addr);
            sym2addr[line[11..line.length]] = addr;
        }
    }

    static void loadMapFile(char[] nmfile) {
        File f = new File(nmfile);
        char[] line;
        RegExp reg = new RegExp(` [\dABCDEF]{4}:[\dABCDEF]{8}\s{7}(.*)`, ``);

        while (!f.eof()) {
            line = f.readLine();
            if (reg.test(line) == 0) continue;
            line = reg.replace(`$1`);
            char[][] seps = std.string.split(line);
            if (seps.length != 2) continue;
            void* addr;
            sscanf(toStringz(seps[1]), "%x", &addr);
            sym2addr[seps[0]] = addr;
        }
    }

private:
    static void*[char[]] sym2addr;
    static char[][] syms;

}

class Class {
public:
    char[] sym;
    char[] name;
    void* address;
    Method[] methods;

    static Class forName(char[] name) {
        if (!(name in classes)) return null;
        return classes[name];
    }

    static Class[char[]] classes;
}

class Method {
    char[] sym;
    char[] fullname;
    char[] name;
    Type type;
    Class klass;
    void* address;
}

class Function {
    // this name 'FQual' avoid a bug of dmd.
    enum FQual {
        STATIC = 0x01,
        DELEGATE = 0x02
    }
    FQual qual;
    Type ret;
    Type[] args;
}

class Builtin {
    enum Type {
        BYTE, UBYTE, SHORT, USHORT, INT, UINT, LONG, ULONG, BIT,
        CHAR, WCHAR, DCHAR,
        FLOAT, DOUBLE, REAL, IFLOAT, IDOUBLE, IREAL,
        CFLOAT, CDOUBLE, CREAL, VOID,
        ARGS, ERR
    }
    Type type;

    static this() {
        simpleMangle['v'] = Type.VOID;
        simpleMangle['g'] = Type.BYTE;
        simpleMangle['h'] = Type.UBYTE;
        simpleMangle['s'] = Type.SHORT;
        simpleMangle['t'] = Type.USHORT;
        simpleMangle['i'] = Type.INT;
        simpleMangle['k'] = Type.UINT;
        simpleMangle['l'] = Type.LONG;
        simpleMangle['m'] = Type.ULONG;
        simpleMangle['f'] = Type.FLOAT;
        simpleMangle['d'] = Type.DOUBLE;
        simpleMangle['e'] = Type.REAL;

        simpleMangle['o'] = Type.IFLOAT;
        simpleMangle['p'] = Type.IDOUBLE;
        simpleMangle['j'] = Type.IREAL;
        simpleMangle['q'] = Type.CFLOAT;
        simpleMangle['r'] = Type.CDOUBLE;
        simpleMangle['c'] = Type.CREAL;

        simpleMangle['b'] = Type.BIT;
        simpleMangle['a'] = Type.CHAR;
        simpleMangle['u'] = Type.WCHAR;
        simpleMangle['w'] = Type.DCHAR;

        simpleMangle['@'] = Type.ERR;
    }

    this(Type t) { type = t; }
    this(char t) {
        if (! (t in simpleMangle)) {
            throw new Error("cannot demangle builtin symbol");
        }
        type = simpleMangle[t];
    }

    static Type[char] simpleMangle;
}

class Type {
    enum Qual {
        ARRAY = 0x01,
        SARRAY = 0x02,
        ASSOC = 0x04,
        POINTER = 0x08,
        REFERENCE = 0x10,
        INOUT = 0x20,
        OUT = 0x40
    }
    Qual qual;
    Type assoc;
    enum Kind {
        UNKNOWN, BUILTIN, KLASS, FUNC
    }
    Kind kind;
    union {
        Builtin builtin;
        Class klass;
        Function func;
        char[] unknown;
    }

    char[] toString() {
        return "";
    }
}

extern (C) void* _d_newclass(void*);
extern (C) void _d_delclass(void**);
